/*
 *      xmms2.c
 *
 *      Copyright 2008 PCMan <pcman.tw@gmail.com>
 *      Copyright 2009 Jürgen Hötzel <juergen@archlinux.org>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <glib/gi18n-lib.h>
#include <glib/gstdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <xmmsclient/xmmsclient.h>
#include <xmmsclient/xmmsclient-glib.h>

#include "utils.h"

#include <lxpanel/plugin.h>

typedef struct {
    unsigned int timer;
    Plugin * plugin;
    int title_size;
    gboolean show_time;
    int volume_val;

    gboolean con_init;
    xmmsc_connection_t *con;
    gpointer gmain_data;

    GtkWidget *box;
    GtkWidget *play_btn;
    GtkWidget *progress_btn;
    GtkWidget *volume_btn;
    GtkWidget *title_label;
    int32_t playback_status;
    uint32_t play_time;
    int32_t cur_track_duration;
    int32_t cur_track_id;
} Xmms2;

typedef struct _TrackProperties{
    const char *artist;
    const char *album;
    const char *title;
    const char *url;
    const char *mime;
    const char *comment;
    int32_t duration;
    int32_t isvbr;
    int32_t bitrate;
    int32_t size;
}TrackProperties;

static void panel_config_changed(Plugin* p);

static void setup_xmms_callbacks(Xmms2* xmms2);

static void on_play_btn_clicked(GtkButton* btn, gpointer user_data);
static gboolean get_track_properties         (xmmsv_t *value, TrackProperties *properties);

inline static gboolean ensure_connection(Xmms2* xmms2, gboolean start_xmms2d)
{
    if(xmms2->con_init && xmms2->con)
        return TRUE;
    if(G_UNLIKELY(!xmms2->con))
        xmms2->con = xmmsc_init ("lxpanel-xmms2");
    if(G_LIKELY(xmms2->con))
    {
        /* try to connect to xmms2d, and launch the daemon if needed. */
        if( !xmmsc_connect(xmms2->con, getenv("XMMS_PATH")))
        {
            if( ! start_xmms2d
                || ! g_spawn_command_line_sync( "xmms2-launcher --yes-run-as-root", NULL, NULL, NULL, NULL )
                || !xmmsc_connect(xmms2->con, getenv ("XMMS_PATH")) )
            {
                return FALSE;
            }
        }
        xmms2->gmain_data = xmmsc_mainloop_gmain_init(xmms2->con);
        /* register callbacks */
        setup_xmms_callbacks(xmms2);
        xmms2->con_init = TRUE;
    }
    return TRUE;
}

void on_progress_bar_changed(GtkScaleButton* btn, gdouble value, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    xmmsc_result_t* res;
    uint32_t new_play_time = value * xmms2->cur_track_duration / 100;
    if(!ensure_connection(xmms2, TRUE))
        return;
    res = xmmsc_playback_seek_ms( xmms2->con, new_play_time, XMMS_PLAYBACK_SEEK_SET );
    xmmsc_result_unref(res);
}

void on_prev_btn_clicked(GtkButton* btn, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    if(!ensure_connection(xmms2, TRUE))
        return;
    xmmsc_result_t* res = xmmsc_playlist_set_next_rel(xmms2->con, -1);
    xmmsc_result_unref(res);
    res = xmmsc_playback_tickle(xmms2->con);
    xmmsc_result_unref(res);
}

void on_next_btn_clicked(GtkButton* btn, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    if(!ensure_connection(xmms2, TRUE))
        return;
    xmmsc_result_t* res = xmmsc_playlist_set_next_rel(xmms2->con, 1);
    xmmsc_result_unref(res);
    res = xmmsc_playback_tickle(xmms2->con);
    xmmsc_result_unref(res);
}

void on_play_btn_clicked(GtkButton* btn, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    xmmsc_result_t *res;
    if(!ensure_connection(xmms2, TRUE))
        return;
    if( xmms2->playback_status == XMMS_PLAYBACK_STATUS_PLAY )
        res = xmmsc_playback_pause(xmms2->con);
    else
        res = xmmsc_playback_start(xmms2->con);
    xmmsc_result_unref(res);
}

void on_stop_btn_clicked(GtkButton* btn, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    if(!ensure_connection(xmms2, TRUE))
        return;
    xmmsc_result_t* res = xmmsc_playback_stop(xmms2->con);
    xmmsc_result_unref(res);
}

static gboolean get_track_properties (xmmsv_t *value, TrackProperties *properties)
{
    /* traverse the dict of dict */
    xmmsv_dict_iter_t *parent_it;

    /* fallback for title: online streams often provide channel */
    const char* channel = NULL;

    /* default values: empty */
    bzero( properties, sizeof(TrackProperties) );

    xmmsv_get_dict_iter (value, &parent_it);
    while (xmmsv_dict_iter_valid (parent_it ) )
    {
        const char *key;
        const char **val_str = NULL;
        int32_t *val_int = NULL;
        xmmsv_t *child_value;
        xmmsv_dict_iter_t *child_it;

        /* get child dict */
        xmmsv_dict_iter_pair (parent_it, &key, &child_value);

        /* check type of property */
        if (strcmp( key, "artist" ) == 0)
            val_str = &(properties->artist);
        else if (strcmp( key, "album" ) == 0)
            val_str = &(properties->album);
        else if (strcmp( key, "mime" ) == 0)
            val_str = &(properties->mime);
        else if (strcmp( key, "comment" ) == 0)
            val_str = &(properties->comment);
        else if (strcmp( key, "channel" ) == 0)
            val_str = &channel;
        else if (strcmp( key, "url" ) == 0)
            val_str = &(properties->url);
        else if (strcmp( key, "title" ) == 0)
            val_str = &(properties->title);
        else if (strcmp( key, "duration" ) == 0)
            val_int = &(properties->duration);
        else if (strcmp( key, "isvbr" ) == 0)
            val_int = &(properties->isvbr);
        else if (strcmp( key, "bitrate" ) == 0)
            val_int = &(properties->bitrate);
        else if (strcmp( key, "size" ) == 0)
            val_int = &(properties->size);

        if (xmmsv_get_dict_iter (child_value, &child_it) &&
            xmmsv_dict_iter_valid (child_it) && (val_int || val_str) &&
            xmmsv_dict_iter_pair (child_it, NULL, &child_value)) {

            if (val_int != NULL)
                xmmsv_get_int( child_value, val_int);
            else
                xmmsv_get_string( child_value, val_str );
        }
        xmmsv_dict_iter_next (parent_it);
    }

    if ((properties->title == NULL) || g_str_equal( properties->title, "" ))
    {
        if (channel == NULL)
            return FALSE;
        else
            properties->title = channel;
    }
    return TRUE;
}

#if 0
static void queue_update_track( uint32_t id, GtkTreeIter* it )
{
    UpdateTrack* ut;
    xmmsc_result_t *res;
    ut = g_slice_new(UpdateTrack);
    ut->id = id;
    ut->it = *it;
    g_hash_table_insert( update_tracks, ut, ut );
    res = xmmsc_medialib_get_info( xmms2->con, id );
    xmmsc_result_notifier_set_full( res, (xmmsc_result_notifier_t)update_track, ut, xmms2 );
    xmmsc_result_unref( res );
}
#endif


static void on_server_disconnect (void *user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    xmmsc_mainloop_gmain_shutdown(xmms2->con, xmms2->gmain_data);
    g_warning( "Server gone" );
    xmmsc_unref (xmms2->con);
    xmms2->con = NULL;
}

static int on_playback_status_changed( xmmsv_t *value, void *user_data )
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    GtkWidget* img;
    if ( !xmmsv_get_int(value, &xmms2->playback_status) )
    {
        xmms2->playback_status = XMMS_PLAYBACK_STATUS_STOP;
        return TRUE;
    }

    switch( xmms2->playback_status )
    {
        case XMMS_PLAYBACK_STATUS_PLAY:
        {
            xmmsc_result_t* res2;
            gtk_widget_set_tooltip_text( xmms2->play_btn, _("Pause") );
            img = gtk_bin_get_child( (GtkBin*)xmms2->play_btn );
            gtk_image_set_from_stock( (GtkImage*)img, GTK_STOCK_MEDIA_PAUSE, GTK_ICON_SIZE_SMALL_TOOLBAR );
            break;
        }
        case XMMS_PLAYBACK_STATUS_STOP:
            gtk_button_set_label( GTK_BUTTON(xmms2->progress_btn), "--:--" );
            gtk_scale_button_set_value( GTK_SCALE_BUTTON(xmms2->progress_btn), 0.0 );
            gtk_label_set_text(GTK_LABEL(xmms2->title_label), _("Stopped"));
        case XMMS_PLAYBACK_STATUS_PAUSE:
            gtk_widget_set_tooltip_text( xmms2->play_btn, _("Play") );
            img = gtk_bin_get_child( (GtkBin*)xmms2->play_btn );
            gtk_image_set_from_stock( (GtkImage*)img, GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_SMALL_TOOLBAR );
            break;
    }
    return TRUE;
}

static int on_playback_playtime_changed( xmmsv_t* value, void* user_data )
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    int32_t time;
    char buf[32];
    if ( xmmsv_is_error(value)
         || ! xmmsv_get_int(value, &time))
        return TRUE;

    time /= 1000;
    if( time == xmms2->play_time )
        return TRUE;
    xmms2->play_time = time;

    /* FIXME: Only set the callback when show_time = TRUE */
    if(xmms2->show_time)
        gtk_button_set_label( GTK_BUTTON(xmms2->progress_btn), timeval_to_str( time, buf, G_N_ELEMENTS(buf) ) );

    if( xmms2->cur_track_duration > 0 )
    {
        g_signal_handlers_block_by_func(xmms2->progress_btn, on_progress_bar_changed, user_data);
        gtk_scale_button_set_value( GTK_SCALE_BUTTON(xmms2->progress_btn),
                              (((gdouble)100000 * time) / xmms2->cur_track_duration) );
        g_signal_handlers_unblock_by_func(xmms2->progress_btn, on_progress_bar_changed, user_data);
    }
    return TRUE;
}

static int on_playback_track_loaded( xmmsv_t* value, void* user_data )
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    TrackProperties track_properties;
    const char* err;
    const char *guessed_title;
    GString* window_title;

    if (xmmsv_get_error (value, &err)) {
        g_warning( "Server error: %s", err );
        return TRUE;
    }

    if (!get_track_properties( value, &track_properties))
        track_properties.title = guessed_title = guess_title_from_url( track_properties.url );
    xmms2->cur_track_duration = track_properties.duration;

    window_title = create_window_title(track_properties.artist, track_properties.title, xmms2->playback_status == XMMS_PLAYBACK_STATUS_PLAY);

    gtk_label_set_text( GTK_LABEL(xmms2->title_label), window_title->str );
    gtk_widget_set_tooltip_text(GTK_LABEL(xmms2->title_label), window_title->str);

    g_string_free( window_title, TRUE );
    return TRUE;
}

static int on_playback_cur_track_changed( xmmsv_t* value, void* user_data )
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    if( xmmsv_get_int(value, &xmms2->cur_track_id) && xmms2->cur_track_id != 0)
    {
        xmmsc_result_t *res2;
        res2 = xmmsc_medialib_get_info(xmms2->con, xmms2->cur_track_id);
        xmmsc_result_notifier_set_and_unref(res2, on_playback_track_loaded, xmms2);
    }

    return TRUE;
}

static void get_channel_volumes( const char *key, xmmsv_t *value, void *user_data)
{
    GSList** volumes = (GSList**)user_data;
    int32_t volume;
    xmmsv_get_int( value , &volume );
    *volumes = g_slist_prepend(*volumes, (gpointer)volume);
}

static void get_channel_volume_names(const char* key, xmmsv_t *value, void *user_data)
{
    GSList** volumes = (GSList**)user_data;
    *volumes = g_slist_prepend(*volumes, (gpointer)key);
}

static int on_volume_btn_set_volume(xmmsv_t *value, void* user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    GSList* volumes = NULL, *l;

    xmmsv_dict_foreach( value, get_channel_volume_names, &volumes);

    for( l = volumes; l; l = l->next )
    {
        xmmsc_result_t* res2;
        res2 = xmmsc_playback_volume_set(xmms2->con, (char*)l->data, xmms2->volume_val);
        xmmsc_result_unref(res2);
    }
    g_slist_free(volumes);
    return TRUE;
}

static void on_volume_btn_changed(GtkScaleButton* btn, gdouble val, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    xmmsc_result_t* res;
    if(!ensure_connection(xmms2, TRUE))
        return;
    res = xmmsc_playback_volume_get(xmms2->con);
    xmms2->volume_val = (int)val;
    xmmsc_result_notifier_set_and_unref(res, on_volume_btn_set_volume, xmms2);
}

static void on_volume_btn_scrolled(GtkWidget *widget, GdkEventScroll *event, gpointer user_data)
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    guint volume;
    GtkAdjustment *vol_adj;
    xmmsc_result_t *res;
    res = xmmsc_playback_volume_get(xmms2->con);
    if (event->type != GDK_SCROLL)
            return;
    vol_adj = gtk_scale_button_get_adjustment (GTK_SCALE_BUTTON(xmms2->volume_btn));
    switch (event->direction)
    {
        case GDK_SCROLL_UP:
                volume = gtk_adjustment_get_value (vol_adj) + 2;
                gtk_adjustment_set_value (GTK_ADJUSTMENT(vol_adj), volume);
                break;
        case GDK_SCROLL_DOWN:
                volume = gtk_adjustment_get_value (vol_adj) - 2;
                gtk_adjustment_set_value (GTK_ADJUSTMENT(vol_adj), volume);
                break;
        default:
                return;
    }
    xmms2->volume_val = volume;
    xmmsc_result_notifier_set_and_unref(res, on_volume_btn_set_volume, xmms2);
}

static int on_playback_volume_changed( xmmsv_t* value, void* user_data )
{
    Xmms2* xmms2 = (Xmms2*)user_data;
    GSList* volumes = NULL, *l;
    uint32_t vol = 0;

    if(xmmsv_is_error ( value ))
    {
        const gchar *error_msg;
        xmmsv_get_error( value, &error_msg );
        g_warning( "%s: %s", __func__, error_msg );
    }
    else if (xmmsv_is_type( value, XMMSV_TYPE_DICT ))
    {
        xmmsv_dict_foreach( value, get_channel_volumes, &volumes );
        for( l = volumes; l; l = l->next )
        {
            if( vol < GPOINTER_TO_UINT(l->data) )
                vol = GPOINTER_TO_UINT(l->data);
        }
        g_slist_free(volumes);

        g_signal_handlers_block_by_func( xmms2->volume_btn, on_volume_btn_changed, NULL );
        gtk_scale_button_set_value( GTK_SCALE_BUTTON(xmms2->volume_btn), vol );
        g_signal_handlers_unblock_by_func( xmms2->volume_btn, on_volume_btn_changed, NULL );
    }
    return TRUE;
}

void setup_xmms_callbacks(Xmms2* xmms2)
{
    xmmsc_result_t* res;

    xmmsc_disconnect_callback_set (xmms2->con, on_server_disconnect, xmms2);
    /* play status */
    res = xmmsc_playback_status(xmms2->con);
    xmmsc_result_notifier_set_and_unref(res, on_playback_status_changed, xmms2);
    XMMS_CALLBACK_SET( xmms2->con, xmmsc_broadcast_playback_status, on_playback_status_changed, xmms2 );

    /* play time */
    /* this is a signal rather than broadcast, so restart is needed in the callback func. */
    XMMS_CALLBACK_SET( xmms2->con, xmmsc_signal_playback_playtime, on_playback_playtime_changed, xmms2);

    /* current track info */
    res = xmmsc_playback_current_id( xmms2->con );
    xmmsc_result_notifier_set_and_unref( res, on_playback_cur_track_changed, xmms2 );
    XMMS_CALLBACK_SET( xmms2->con, xmmsc_broadcast_playback_current_id, on_playback_cur_track_changed, xmms2 );

    /* volume */
    res = xmmsc_playback_volume_get(xmms2->con);
    xmmsc_result_notifier_set_and_unref(res, on_playback_volume_changed, xmms2 );
    XMMS_CALLBACK_SET( xmms2->con, xmmsc_broadcast_playback_volume_changed, on_playback_volume_changed, xmms2 );
}

static GtkWidget* add_btn(Xmms2* xmms2, char* icon, const char* tooltip, GCallback cb)
{
    GtkWidget* btn = gtk_button_new();
    GtkWidget* img = gtk_image_new_from_stock(icon, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_button_set_relief(btn, GTK_RELIEF_NONE);
    GTK_WIDGET_UNSET_FLAGS(btn, GTK_CAN_FOCUS);
    gtk_misc_set_padding(GTK_MISC(img), 0, 0);
    gtk_container_set_border_width(GTK_CONTAINER(btn), 0);
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_set_tooltip_text(btn, tooltip);
    g_signal_connect(btn, "clicked", cb, xmms2);
    gtk_box_pack_start(GTK_BOX(xmms2->box), btn, FALSE, TRUE, 0);
    return btn;
}

static void setup_ui(Plugin* plugin, Xmms2* xmms2)
{
    GtkWidget *btn, *img, *switch_pl_mi, *show_pl_mi;
    xmmsc_result_t* res;

    xmms2->box = gtk_hbox_new(FALSE, 0);

    add_btn(xmms2, GTK_STOCK_MEDIA_PREVIOUS, _("Previous"), G_CALLBACK(on_prev_btn_clicked));
    xmms2->play_btn = add_btn(xmms2, GTK_STOCK_MEDIA_PLAY, _("Play"), G_CALLBACK(on_play_btn_clicked));
    add_btn(xmms2, GTK_STOCK_MEDIA_STOP, _("Stop"), G_CALLBACK(on_stop_btn_clicked));
    add_btn(xmms2, GTK_STOCK_MEDIA_NEXT, _("Next"), G_CALLBACK(on_next_btn_clicked));

    xmms2->title_label = gtk_label_new(_("Stopped"));
    gtk_label_set_ellipsize(GTK_LABEL(xmms2->title_label), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(xmms2->box), xmms2->title_label, TRUE, TRUE, 0);

    xmms2->progress_btn = gtk_scale_button_new(GTK_ICON_SIZE_INVALID, 0, 100, 1, NULL);
    GTK_WIDGET_UNSET_FLAGS(xmms2->progress_btn, GTK_CAN_FOCUS);
    gtk_button_set_label(GTK_BUTTON(xmms2->progress_btn), "--:--");
    gtk_scale_button_set_orientation(GTK_SCALE_BUTTON(xmms2->progress_btn), GTK_ORIENTATION_HORIZONTAL);
    gtk_box_pack_start(GTK_BOX(xmms2->box), xmms2->progress_btn, FALSE, TRUE, 0);
    g_signal_connect(xmms2->progress_btn, "value-changed", G_CALLBACK(on_progress_bar_changed), xmms2);

    /* add volume button */
    xmms2->volume_btn = gtk_volume_button_new();
    GTK_WIDGET_UNSET_FLAGS(xmms2->volume_btn, GTK_CAN_FOCUS);
    gtk_scale_button_get_adjustment(GTK_SCALE_BUTTON(xmms2->volume_btn))->upper = 100;
    gtk_box_pack_start(GTK_BOX(xmms2->box), xmms2->volume_btn, FALSE, TRUE, 0);
    g_signal_connect(xmms2->volume_btn, "value-changed", G_CALLBACK(on_volume_btn_changed), xmms2);
    g_signal_connect(xmms2->volume_btn, "scroll-event", G_CALLBACK(on_volume_btn_scrolled), xmms2);

    gtk_widget_show_all(xmms2->box);
    plugin->pwid = xmms2->box;

    ensure_connection(xmms2, FALSE);
}

static int xmms2_constructor(Plugin *p, char** fp)
{
    Xmms2 *xmms2;
    char buffer [60];

#ifdef ENABLE_NLS
    static gboolean init_gettext = FALSE;
    if(!init_gettext)
    {
        bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
        bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
        init_gettext = TRUE;
    }
#endif

    ENTER;

    xmms2 = g_new0(Xmms2, 1);
    xmms2->plugin = p;
    p->priv = xmms2;
    xmms2->show_time = TRUE;
    xmms2->title_size = 48;

    /* build the GUI */
    setup_ui(p, xmms2);

    line s;
    s.len = 256;
    if (fp)
    {
        while (lxpanel_get_line(fp, &s) != LINE_BLOCK_END)
        {
            if (s.type == LINE_NONE) {
                ERR( "xmms2: illegal token %s\n", s.str);
                goto error;
            }
            if (s.type == LINE_VAR) {
                if (!g_ascii_strcasecmp(s.t[0], "title_size")){
                    xmms2->title_size = atoi(s.t[1]);
                }else if (!g_ascii_strcasecmp(s.t[0], "show_time")){
                    xmms2->show_time = atoi(s.t[1]);
                }else {
                    ERR( "example: unknown var %s\n", s.t[0]);
                    continue;
                }
            }
            else {
                ERR( "example: illegal in this context %s\n", s.str);
                goto error;
            }
        }
    }

    panel_config_changed(p);
    gtk_widget_show(p->pwid); /* show plugin on panel */

    RET(TRUE);
error:
    destructor( p );
    RET(FALSE);
}

void panel_config_changed(Plugin* p)
{
    ENTER;
    Xmms2 *xmms2 = (Xmms2 *) p->priv;
    gtk_orientable_set_orientation(xmms2->box, p->panel->orientation);
    if(p->panel->orientation == ORIENT_HORIZ)
    {
        if(xmms2->show_time)
            gtk_widget_show(xmms2->progress_btn);
        gtk_widget_show(xmms2->title_label);
        gtk_widget_set_size_request(xmms2->title_label, xmms2->title_size, -1);
    }
    else
    {
        gtk_widget_hide(xmms2->progress_btn);
        gtk_widget_hide(xmms2->title_label);
    }
    RET();
}

static void config(Plugin *p, GtkWindow* parent) {
    ENTER;
    RET();
}

static void xmms2_destructor(Plugin *p)
{
    ENTER;
    Xmms2 *xmms2 = (Xmms2 *)p->priv;
    xmmsc_unref (xmms2->con);
    /*
    this is not needed as plugin_delete will call gtk_destroy().
    gtk_widget_destroy(xmms2->box);
    */
    g_free(xmms2);
    RET();
}

static void save_config( Plugin* p, FILE* fp )
{
    ENTER;
    Xmms2 *xmms2 = (Xmms2 *)p->priv;
    RET();
}

PluginClass xmms2_plugin_class = {

    PLUGINCLASS_VERSIONING,

    type : "xmms2",
    name : N_("XMMS2 Music Player Controller"),
    version: "0.1",
    description : N_("Control xmms2 music player"),

    constructor : xmms2_constructor,
    destructor  : xmms2_destructor,
    config      : config,
    save        : save_config,
    panel_configuration_changed : panel_config_changed
};
